"""Dashboard powered by Plotly Dash."""


import locale

import dash
import dash_html_components as html
import yaml
from sqlalchemy import create_engine

from .plot import Box, Line, Pie

locale.setlocale(locale.LC_TIME, "")


DEFAULT_TITLE = "HiBro: Home Assistant History Browser"


class HiBro:
    """Base class for the HiBro app."""

    def __init__(self, config_file: str):
        self.config_file = config_file
        self.config = self.read_config()

    @property
    def engine(self):
        return create_engine(self.config["db_url"])

    def read_config(self) -> dict:
        """Read the config file."""
        with open(self.config_file, "r") as f:
            d = yaml.safe_load(f)
        return d

    def get_layout(self):
        """Get the Dash layout."""
        elements = []
        layout = self.config.get("elements")
        for element in layout:
            typ = element.get("type")
            if typ == "line":
                elements += [Line(engine=self.engine, config=element).graph()]
            elif typ == "box":
                elements += [Box(engine=self.engine, config=element).graph()]
            elif typ == "pie":
                elements += [Pie(engine=self.engine, config=element).graph()]
            else:
                raise ValueError(f"Element type {typ} not recognized.")
        return html.Div(
            [
                html.Div(
                    [
                        html.Div(
                            [html.H1(self.config.get("title", DEFAULT_TITLE))],
                            className="container",
                        )
                    ],
                    id="title",
                ),
                html.Div(elements, className="container"),
            ]
        )


EXTERNAL_STYLESHEETS = []


def create_app(config_file: str):
    """Dash app factory."""
    app = dash.Dash(
        __name__,
        meta_tags=[{"name": "viewport", "content": "width=device-width"},],
        external_stylesheets=EXTERNAL_STYLESHEETS,
    )

    app.config.suppress_callback_exceptions = True

    hibro_inst = HiBro(config_file)
    app.title = hibro_inst.config.get("title", DEFAULT_TITLE)
    app.layout = hibro_inst.get_layout

    return app.server



"""Plot functions."""

from datetime import timedelta

import plotly.graph_objects as go
import dash_core_components as dcc
import voluptuous as vol

from .data import get_data

DEFAULT_DURATION = timedelta(hours=24)


def coerce_entity_dict(v):
    """Coerce entity ID strings to a dictionary with key "entity"."""
    if isinstance(v, str):
        return {"entity": v}
    return v


def coerce_timedelta(v):
    """Coerce a dictionary to a timedelta."""
    if isinstance(v, timedelta):
        return v
    elif isinstance(v, dict):
        return timedelta(**v)
    else:
        raise ValueError(f"Duration {v} not understood")


class Plot:
    """Base class for plots."""

    CONFIG_SCHEMA = None

    def __init__(self, engine, config: dict):
        """Initialize given an SQLAlchemy engine and a config dict."""
        self.engine = engine
        self.raw_config = config
        self.config = self.CONFIG_SCHEMA(self.raw_config)

    def data_entity(self, entity_config):
        """Get data for a single entity."""
        entity = entity_config["entity"]
        attribute = entity_config.get("attribute")
        duration = entity_config["duration"]
        resample = entity_config.get("resample")
        aggregate = entity_config.get("aggregate")
        data = get_data(
            engine=self.engine,
            entity=entity,
            attribute=attribute,
            duration=duration,
            resample=resample,
            aggregate=aggregate,
        ) or {}
        name = data.get("name") or entity
        if resample:
            name = f"{name} ({resample}, {aggregate or 'mean'})"
        elif aggregate:
            name = f"{name} ({aggregate or 'ptp'})"
        unit = data.get("unit")
        return name, unit, data


class Line(Plot):
    """Line chart."""

    CONFIG_SCHEMA = vol.Schema(
        {
            vol.Required("type"): "line",
            vol.Required("entities"): [
                vol.All(
                    coerce_entity_dict,
                    vol.Schema(
                        {
                            vol.Required("entity"): str,
                            vol.Optional(
                                "duration", default=DEFAULT_DURATION
                            ): coerce_timedelta,
                            vol.Optional("attribute"): str,
                            vol.Optional("resample"): str,
                            vol.Optional("aggregate"): str,
                        }
                    ),
                )
            ],
        }
    )

    def trace(self, figure, entity_config: dict):
        """Add a single Plotly trace for one entity to the `figure`."""
        name, unit, data = self.data_entity(entity_config)
        if data != {}:
            figure.add_trace(
                go.Scatter(
                    x=data["data"]["time"],
                    y=data["data"][entity_config["entity"]],
                    mode="lines",
                    line_shape="hv",
                    name=name,
                )
            )
        return name, unit

    def graph(self) -> dcc.Graph:
        """Return the dash layout."""
        fig = go.Figure()
        unit = None
        for entity_config in self.config["entities"]:
            _, _unit = self.trace(fig, entity_config)
            unit = _unit or unit
        fig.update_layout(
            xaxis_title=None,
            yaxis_title=unit,
            legend=dict(x=0, y=-0.2, orientation="h"),
            margin=dict(t=10),
            font=dict(family="Roboto", color="#333333"),
            template="plotly_white",
            height=300,
        )
        return dcc.Graph(figure=fig)


class Box(Plot):
    """Box plot."""

    CONFIG_SCHEMA = vol.Schema(
        {
            vol.Required("type"): "box",
            vol.Required("entities"): [
                vol.All(
                    coerce_entity_dict,
                    vol.Schema(
                        {
                            vol.Required("entity"): str,
                            vol.Optional(
                                "duration", default=DEFAULT_DURATION
                            ): coerce_timedelta,
                            vol.Optional("attribute"): str,
                        }
                    ),
                )
            ],
        }
    )

    def trace(self, figure, entity_config: dict):
        """Add a single Plotly trace for one entity to the `figure`."""
        name, unit, data = self.data_entity(entity_config)
        if data != {}:
            figure.add_trace(go.Box(y=data["data"][entity_config["entity"]], name=name,))
        return name, unit

    def graph(self) -> dcc.Graph:
        """Return the dash layout."""
        fig = go.Figure()
        unit = None
        for entity_config in self.config["entities"]:
            _, _unit = self.trace(fig, entity_config)
            unit = _unit or unit
        fig.update_layout(
            xaxis_title=None,
            yaxis_title=unit,
            legend=dict(x=0, y=-0.2, orientation="h"),
            margin=dict(t=10),
            font=dict(family="Roboto", color="#333333"),
            template="plotly_white",
            height=300,
        )
        return dcc.Graph(figure=fig)


class Pie(Plot):
    """Pie chart."""

    CONFIG_SCHEMA = vol.Schema(
        {
            vol.Required("type"): "pie",
            vol.Required("entities"): [
                vol.All(
                    coerce_entity_dict,
                    vol.Schema(
                        {
                            vol.Required("entity"): str,
                            vol.Optional(
                                "duration", default=DEFAULT_DURATION
                            ): coerce_timedelta,
                            vol.Optional("attribute"): str,
                            vol.Optional("aggregate", default="ptp"): str,
                        }
                    ),
                )
            ],
        }
    )

    def piece(self, entity_config: dict):
        """Get label and value for a single piece of the pie."""
        name, unit, data = self.data_entity(entity_config)
        return name, data

    def graph(self) -> dcc.Graph:
        """Return the dash layout."""
        labels = []
        values = []
        for entity_config in self.config["entities"]:
            name, data = self.piece(entity_config)
            if data != {}:
                labels.append(name)
                values.append(data["data"][0][0])
        fig = go.Figure(data=[go.Pie(labels=labels, values=values)])
        fig.update_layout(
            font=dict(family="Roboto", color="#333333"),
            template="plotly_white",
            margin=dict(t=30, b=30),
            height=300,
        )
        return dcc.Graph(figure=fig)



"""Module to fetch history data."""

import json
from datetime import datetime, timedelta
from typing import Optional

import numpy as np
import pandas as pd
import tzlocal
from sqlalchemy import (
    Column,
    DateTime,
    ForeignKey,
    Integer,
    String,
    Text,
)
from sqlalchemy.engine.base import Engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm.session import Session

# SQLAlchemy Schema
# pylint: disable=invalid-name
Base = declarative_base()


class States(Base):  # type: ignore
    """State change history SQLAlchemy model.

    Directly taken from homeassistant."""

    __tablename__ = "states"
    state_id = Column(Integer, primary_key=True)
    domain = Column(String(64))
    entity_id = Column(String(255), index=True)
    state = Column(String(255))
    attributes = Column(Text)
    event_id = Column(Integer, ForeignKey("events.event_id"), index=True)
    last_changed = Column(DateTime(timezone=True), default=datetime.utcnow)
    last_updated = Column(DateTime(timezone=True), default=datetime.utcnow, index=True)
    created = Column(DateTime(timezone=True), default=datetime.utcnow)
    context_id = Column(String(36), index=True)
    context_user_id = Column(String(36), index=True)


def get_data(
    engine: Engine,
    entity: str,
    attribute: Optional[str] = None,
    duration: Optional[timedelta] = None,
    resample: Optional[str] = None,
    aggregate: Optional[str] = None,
):
    """Fetch the data from the database and apply the necessary operations."""
    session = Session()
    query = (
        session.query(States.created, States.state, States.attributes)
        .filter_by(entity_id=entity)
        .order_by(States.created)
    )
    if duration is not None:
        query = query.filter(States.created >= datetime.utcnow() - duration)
    df = pd.read_sql_query(query.statement, con=engine)
    if len(df) == 0:
        return None
    attributes = json.loads(df.iloc[0]["attributes"])
    if attribute is not None:
        df.loc[:, "attributes"] = df.loc[:, "attributes"].map(
            lambda s: json.loads(s).get(attribute)
        )
        df = df.drop("state", axis=1)
    else:
        df = df.drop("attributes", axis=1)
    df.columns = ["time", entity]
    # drop rows with unkown state
    df = df[(df[entity] != "unknown") & (df[entity] != "Invalid")]
    if resample is not None:
        df = df.set_index("time")
        df = (
            df.astype(float)
            .resample("s")
            .pad()
            .resample(resample)
            .agg(aggregate or "mean")
        )
        df = df.reset_index()
    elif aggregate is not None:
        if aggregate == "ptp":
            agg = np.ptp
        else:
            agg = aggregate
        df = df.set_index("time")
        df = df.astype(float).resample("s").agg(agg)
        df = df.reset_index()
    if df.loc[:, "time"].dt.tz is None:
        # localize to UTC if tz-naive
        df.loc[:, "time"] = df.loc[:, "time"].dt.tz_localize("UTC")
    df.loc[:, "time"] = df.loc[:, "time"].dt.tz_convert(tzlocal.get_localzone())
    return {
        "name": attributes.get("friendly_name"),
        "unit": attributes.get("unit_of_measurement"),
        "data": df,
    }



#!/usr/bin/env python

"""The setup script."""

from setuptools import find_packages, setup

with open("README.md") as f:
    README = f.read()


setup(
    author="David Straub",
    author_email="straub@protonmail.com",
    python_requires=">=3.6",
    url="https://github.com/DavidMStraub/HiBro",
    description="Standalone history browser for Home Assistant",
    install_requires=[
        "Click>=7.0",
        "dash",
        "pyyaml",
        "sqlalchemy",
        "voluptuous",
        "pandas",
        "tzlocal",
    ],
    long_description=README,
    long_description_content_type="text/markdown",
    include_package_data=True,
    keywords="hibro",
    name="hibro",
    packages=find_packages(include=["hibro", "hibro.*"]),
    entry_points={"console_scripts": ["hibro=hibro.__main__:main"]},
    test_suite="tests",
    version="0.2.1",
    zip_safe=False,
)



"""Console script for hibro."""

import sys
import click

from .dashboard import create_app


@click.command()
@click.option("--port", default=8456, help="port")
@click.option("--config-file", default="hibro-config.yaml", help="configuration file")
def main(port, config_file):
    """Run the HiBro dashboard."""
    app = create_app(config_file=config_file)
    app.run(port=port, debug=True)


if __name__ == "__main__":
    main()
